
// $Id: AtomData.cpp,v 1.14 2014/05/01 16:46:27 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AtomData.h"
#include "Debug.h"

#include "XcbProperties.h"
#include "XcbUtil.h"

#include <QStringList>
#include <QTextCodec>

//_____________________________________________________________________
AtomData::~AtomData( void )
{
    #if HAVE_X11
    if( data_ ) ::free( data_ );
    #endif
}

//_____________________________________________________________________
void AtomData::setData( void* value, int length )
{
    if( data_ ) ::free( data_ );
    data_ = new unsigned char[length];
    memcpy( data_, value, length );
}

//_____________________________________________________________________
void AtomData::parse( XcbProperties* xcbProperties )
{

    // do nothing for invalid atoms
    if( !isValid() ) return;

    // parse based on type
    if( type_ == "STRING" )
    {

        QTextCodec* codec( QTextCodec::codecForLocale() );
        if( !codec ) return;
        int offset = 0;
        QStringList values;
        const char* data = reinterpret_cast<const char*>(data_);
        while( true )
        {
            QString value( codec->toUnicode( data + offset ).left( itemCount_ - offset ) );
            values.append( value );
            offset += value.size()+1;
            if( offset >= itemCount_ ) break;
        }

        if( name_ == "WM_CLASS" )
        {

            if( values.size() == 1 ) stringValue_ = values[0];
            else if( values.size() > 1 ) stringValue_ = QString( "%1 (%2)" ).arg( values[0] ).arg( values[1] );

        } else if( name_ == "WM_COMMAND" ) {

            stringValue_ = values.join( " " );

        } else stringValue_ = values.join( ", " );

    } else if( type_ == "UTF8_STRING" ) {

        if( length_ > 0 )
        { stringValue_ = QString::fromUtf8( reinterpret_cast<const char*>(data() ), length_ ); }

    } else if( type_ == "CARDINAL" ) {

        // don't parse icon
        if( name_ == "_NET_WM_ICON" )
        {
            if( format_ != 32 ) return;
            QStringList values;
            int offset( 0 );
            uint32_t* data = reinterpret_cast<uint32_t*>( data_ );

            while( true )
            {
                const int width( (int)data[offset] );
                const int height( (int)data[offset+1] );
                values.append( QString( "%1x%2" ).arg( width ).arg( height ) );
                offset += 2+width * height;
                if( offset >= length_/format_ ) break;
            }

            if( !values.empty() )
            { stringValue_ = QString( tr( "Available Sizes: " ) ) + values.join( ", " ); }

        } else {

            QStringList values;
            if( format_ == 32 )
            {

                uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
                for( int i = 0; i < itemCount_; ++i )
                { values.append( QString::number( data[i] ) ); }

            } else if( format_ == 16 ) {

                uint16_t* data =  reinterpret_cast<uint16_t*>( data_ );
                for( int i = 0; i < itemCount_; ++i )
                { values.append( QString::number( data[i] ) ); }

            }

            stringValue_ = values.join( ", " );

        }

    } else if( type_ == "WINDOW" ) {

        QStringList values;
        if( format_ == 32 )
        {

            uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
            for( int i = 0; i < itemCount_; ++i )
            { values.append( XcbUtil::get().windowIdString( data[i] ) ); }

            stringValue_ = values.join( ", " );

        }

    } else if( type_ == "WM_HINTS" ) {

        // parse wm hints
        if( format_ != 32 ) return;
        QStringList values;
        uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
        if( itemCount_ > 1 ) values.append( QString( tr( "Input Focus: %1" ) ).arg( data[1] ? "True":"False" ) );
        if( itemCount_ > 2 )
        {
            switch( data[2] )
            {
                case 0: values.append( QString( tr( "Initial State: Don't Care" ) ) ); break;
                case 1: values.append( QString( tr( "Initial State: Normal" ) ) ); break;
                case 2: values.append( QString( tr( "Initial State: Zoomed" ) ) ); break;
                case 3: values.append( QString( tr( "Initial State: Iconic" ) ) ); break;
                case 4: values.append( QString( tr( "Initial State: Inactive" ) ) ); break;
                default: break;
            }
        }

        if( itemCount_ > 3 && data[3]) values.append( QString( tr( "Icon Bitmap Id: %1" ) ).arg( XcbUtil::get().windowIdString( data[3] ) ) );
        if( itemCount_ > 4 && data[4] ) values.append( QString( tr( "Icon Window Id: %1" ) ).arg( XcbUtil::get().windowIdString(data[4] ) ) );
        if( itemCount_ > 7 && data[7] ) values.append( QString( tr( "Icon Mask Bitmap Id: %1" ) ).arg( XcbUtil::get().windowIdString( data[7] ) ) );
        if( itemCount_ > 6 && (data[5] || data[6]) ) values.append( QString( tr( "Starting Icon Position: %1, %2" ) ).arg( data[5] ).arg( data[6] ) );
        if( itemCount_ > 8 && data[8]) values.append( QString( tr( "Group Leader Window Id: %1" ) ).arg( XcbUtil::get().windowIdString( data[8] ) ) );

        stringValue_ = values.join( "\n" );

    } else if( type_ == "WM_SIZE_HINTS" ) {

        // parse wm format hints
        if( format_ != 32 ) return;

        QStringList values;
        uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
        if( itemCount_ > 2 && (data[1] != 0 || data[2] != 0) ) values.append( QString( tr( "Specified Location: %1, %2" ) ).arg( data[1] ).arg( data[2] ));
        if( itemCount_ > 4 && (data[3] != 0 || data[4] != 0) ) values.append( QString( tr( "Specified Size: %1x%2" ) ).arg( data[3] ).arg( data[4] ));
        if( itemCount_ > 6 && (data[5] != 0 || data[6] != 0) ) values.append( QString( tr( "Specified Minimum Size: %1x%2" ) ).arg( data[5] ).arg( data[6] ));
        if( itemCount_ > 8 && (data[7] != 0 || data[8] != 0) ) values.append( QString( tr( "Specified Maximum Size: %1x%2" ) ).arg( data[7] ).arg( data[8] ));
        if( itemCount_ > 10 && (data[9] != 0 || data[10] != 0) ) values.append( QString( tr( "Specified Size Increment: %1x%2" ) ).arg( data[9] ).arg( data[10] ));
        if( itemCount_ > 12 && (data[11] != 0 || data[12] != 0) ) values.append( QString( tr( "Specified Minimum Aspect Ratio: %1/%2" ) ).arg( data[11] ).arg( data[12] ));
        if( itemCount_ > 14 && (data[13] != 0 || data[14] != 0) ) values.append( QString( tr( "Specified Maximum Aspect Ratio: %1/%2" ) ).arg( data[13] ).arg( data[14] ));
        if( itemCount_ > 16 && (data[15] != 0 || data[16] != 0) ) values.append( QString( tr( "Specified Base Size: %1x%2" ) ).arg( data[15] ).arg( data[16] )) ;
        if( itemCount_ > 18  )
        {
            switch( data[17] )
            {
                case 0: values.append( QString( tr( "Gravity: Forget" ) )); break;
                case 1: values.append( QString( tr( "Gravity: North-West" ) )); break;
                case 2: values.append( QString( tr( "Gravity: North" ) )); break;
                case 3: values.append( QString( tr( "Gravity: North-East" ) )); break;
                case 4: values.append( QString( tr( "Gravity: West" ) )); break;
                case 5: values.append( QString( tr( "Gravity: Center" ) )); break;
                case 6: values.append( QString( tr( "Gravity: East" ) )); break;
                case 7: values.append( QString( tr( "Gravity: South-West" ) )); break;
                case 8: values.append( QString( tr( "Gravity: South" ) )); break;
                case 9: values.append( QString( tr( "Gravity: South-East" ) )); break;
                case 10: values.append( QString( tr( "Gravity: Static" ) )); break;
                default: break;
            }
        }
        stringValue_ = values.join( "\n" );

    } else if( type_ == "WM_STATE" ) {

        // parse wm state
        if( format_ != 32 ) return;

        QStringList values;
        uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
        if( itemCount_ > 0 )
        {
            switch( data[0] )
            {
                case 0: values.append( QString( tr( "Window State: Withdrawn" ) ) ); break;
                case 1: values.append( QString( tr( "Window State: Normal" ) ) ); break;
                case 3: values.append( QString( tr( "Window State: Iconic" ) ) ); break;
                default: break;
            }
        }

        if( itemCount_ > 1 && data[1] ) values.append( QString( tr( "Icon Window: %1" ) ).arg( XcbUtil::get().windowIdString( data[1] ) ) );
        stringValue_ = values.join( "\n" );

    } else if( type_ == "ATOM" ) {

        #if HAVE_X11
        QStringList values;
        xcb_atom_t *states = reinterpret_cast<xcb_atom_t*>(data_);
        for( int i = 0; i < itemCount_; ++i )
        { values.append( xcbProperties->atomName( states[i] ) ); }

        stringValue_ =  values.join( "\n" );
        #endif

    } else if( format_ == 32 ) {

        QStringList values;
        uint32_t* data = reinterpret_cast<uint32_t*>( data_ );
        for( int i = 0; i < itemCount_; ++i )
        { values.append( XcbUtil::get().windowIdString( data[i] ) ); }

        stringValue_ =  values.join( ", " );

    } else if( format_ == 16 ) {

        QStringList values;
        uint16_t* data = reinterpret_cast<uint16_t*>( data_ );
        for( int i = 0; i < itemCount_; ++i )
        { values.append( XcbUtil::get().windowIdString( data[i] ) ); }

        stringValue_ =  values.join( ", " );

    } else if( format_ == 8 ) {

        QStringList values;
        unsigned char* data = reinterpret_cast<unsigned char*>(data_);
        for( int i = 0; i < itemCount_; ++i )
        { values.append( XcbUtil::get().windowIdString( data[i] ) ); }

        stringValue_ =  values.join( ", " );

    }

}

//________________________________________________________
void AtomData::free( void )
{
    if( data_ ) ::free( data_ );
    data_ = 0;
    itemCount_ = 0;
    length_ = 0;
    format_ = 0;
}
